﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace ScriptEditor
{
    public class ScriptCompile
    {
        public static void Compile(string data, string filename)
        {
            char[] sep = { '(', ')', ',', ';', ':', '{', '}', '[', ']', '!', '+', '-', '*', '/', '%', '=', '<', '>', '|', '&', '^', ' ', '\r', '\n' };
            string[] consts1 = { "(", ")", ",", ";", "{", "}", "[", "]", "!", "+", "-", "*", "/", "%", "==", "!=", "<", ">", "<=", ">=", "=", "&&", "||", "^", "++", "--" };
            string[] consts2 = { "integer", "string", "ust", "location", "object", "list", "void", "if", "else", "while", "for", "continue", "break", "switch", "case", "default", "return", "function", "on", "scriptvar", "include", "extern" };
            // 0x34 // function ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(\((?:[0-9a-zA-Z\,\ ]+|)\))
            //string functionpars = "function ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(\\((?:[0-9a-zA-Z\\,\\ ]+|)\\))";
            // 0x35 // on ([\w]+)(?:(\<[0-9a-zA-Z\"]+\>)|)(\((?:[\w\, ]+|)\))
            //string eventpars = "on ([\\w]+)(?:(\\<[0-9a-zA-Z\\\"]+\\>)|)(\\((?:[\\w\\, ]+|)\\))";
            // 0x36 // scriptvar ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(?:((?:\ |)=(?:\ |)[a-zA-Z\"\ \,\(\)\;]+|\;))
            //string scriptvarpars = "scriptvar ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(?:((?:\\ |)=(?:\\ |)[a-zA-Z\"\\ \\,\\(\\)\\;]+|\\;))";
            // 0x37 // include ([0-9a-zA-Z]+;)
            //string includepars = "include ([0-9a-zA-Z]+)";
            // 0x38 // extern ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(\((?:[0-9a-zA-Z\,\ ]+|)\);)
            //string externpars = "extern ([0-9a-zA-Z]+) ([0-9a-zA-Z]+)(\\((?:[0-9a-zA-Z\\,\\ ]+|)\\);)";
            // 0x3E // (0x[0-9A-F]{8})
            string constint32pars = "(0x[0-9A-F]{8})";
            // 0x3D // (0x[0-9A-F]{4})
            string constshortpars = "(0x[0-9A-F]{4})";
            // 0x3C // (0x[0-9A-F]{2})
            string constbytepars = "(0x[0-9A-F]{2})";

            uint rbracketslevel = 0;
            //uint fbracketslevel = 0;
            string str = String.Empty;
            bool isquote = false;
            bool isevent = false;
            //bool lastwasevent = false;
            char prevchar = ' ';
            List<string> strs = new List<string>();

            for (int i = 0; i < data.Length; ++i)
            {
                char ch = data[i];
                if (ch == '(' && !isquote) ++rbracketslevel;
                if (ch == ')' && !isquote) --rbracketslevel;
                //if (ch == '{' && !isquote) ++fbracketslevel;
                //if (ch == '}' && !isquote) --fbracketslevel;
                if (isevent && rbracketslevel > 0 || isevent && rbracketslevel==0 && ch==')')
                    continue;
                //if (lastwasevent && ch == '<') continue;
                //if (lastwasevent && ch == '>') { ch = ' '; lastwasevent = false; }
                if (consts1.Contains(prevchar.ToString() + ch.ToString()) && (prevchar.ToString() + ch.ToString()).Length==2)
                {
                    strs.RemoveAt(strs.Count-1);
                    str += prevchar.ToString();
                }
                if (sep.Contains(ch) && !isquote && !consts1.Contains((prevchar.ToString()+ch.ToString())))
                {
                    if (str == "on") { isevent = true; /*lastwasevent = true;*/ }
                    if (ch == '{' && isevent) isevent = false;
                    if (isevent && ch == '<') ch = '(';
                    if (isevent && ch == '>') ch = ')';
                    // work with str

                    if (str!=String.Empty)
                        strs.Add(str);
                    if (ch != ' ' && ch != '\r' && ch != '\n' && ch != ':')
                        strs.Add(ch.ToString());
                    str = String.Empty;
                }
                else
                {
                    if (ch == '"') isquote = !isquote;
                    str += ch;
                }
                prevchar = ch;
            }

            StreamWriter writer = new StreamWriter(filename);
            BinaryWriter bwriter = new BinaryWriter(writer.BaseStream);

            NextIs next = NextIs.None;
            uint value = 0;

            foreach (string ou in strs)
            {
                byte b = StringToByte(ou);
                next = NextIs.None;
                if (b == 0xFF)
                {
                    
                    if (Regex.IsMatch(ou, constint32pars)) { next = NextIs.Int32; b = 0x3E; value = Convert.ToUInt32(ou, 16); }
                    else if (Regex.IsMatch(ou, constshortpars)) { next = NextIs.Short; b = 0x3D; value = Convert.ToUInt32(ou, 16); }
                    else if (Regex.IsMatch(ou, constbytepars)) { next = NextIs.Byte; b = 0x3C; value = Convert.ToUInt32(ou,16); }
                    else if (ou.StartsWith("\"") && ou.EndsWith("\"") && SDB.FindLine(ou) > 0) { next = NextIs.QuotedString; b = 0x3A; value = SDB.FindLine(ou); }
                    else if (SDB.FindLine(ou) > 0) { next = NextIs.String; b = 0x41; value = SDB.FindLine(ou); }
                    else //Here we will need to add new string to SDB
                    {
                        if (ou.StartsWith("\"") && ou.EndsWith("\"")) { next = NextIs.QuotedString; b = 0x3A; value = SDB.WriteEntry(ou); }
                        else { next = NextIs.String; b = 0x41; value = SDB.WriteEntry(ou); }
                    }
                }

                bwriter.Write(ByteToObfuscated(b));
                switch(next)
                {
                    case NextIs.Byte: bwriter.Write((byte)value); break;
                    case NextIs.Short: bwriter.Write((short)value); break;
                    case NextIs.Int32: bwriter.Write((uint)value); break;
                    case NextIs.String:
                    case NextIs.QuotedString: bwriter.Write((short)value); break;
                    default: break;
                }
            }
            bwriter.Flush();
            bwriter.Close();
        }

        public enum NextIs
        {
            None,
            String,
            QuotedString,
            Byte,
            Short,
            Int32
        }

        private static byte StringToByte(string obfusbyte)
        {
            switch (obfusbyte.ToLower())
            {
                case "(": return 0x02;
                case ")": return 0x03;
                case ",": return 0x04;
                case ";": return 0x06;
                case "{": return 0x07;
                case "}": return 0x08;
                case "[": return 0x09;
                case "]": return 0x0A;
                case "!": return 0x0B;
                case "+": return 0x0C;
                case "-": return 0x0D;
                case "*": return 0x0E;
                case "/": return 0x0F;
                case "%": return 0x10;
                case "==": return 0x11;
                case "!=": return 0x12;
                case "<": return 0x13;
                case ">": return 0x14;
                case "<=": return 0x15;
                case ">=": return 0x16;
                case "=": return 0x17;
                case "&&": return 0x18;
                case "||": return 0x19;
                case "^": return 0x1A; //(xor)
                case "++": return 0x1B;
                case "--": return 0x1C;
                case "integer": return 0x1D;
                case "string": return 0x1E;
                case "ust": return 0x1F;
                case "location": return 0x20; // location
                case "object": return 0x21;
                case "list": return 0x22;
                case "void": return 0x23;
                case "if": return 0x25;
                case "else": return 0x26;
                case "ENDIF": return 0x27;  //unused
                case "while": return 0x28;
                case "ENDWHILE": return 0x29;   //unused
                case "for": return 0x2A;
                case "ENDFOR": return 0x2B;
                case "continue": return 0x2C;
                case "break": return 0x2D;
                case "GOTO": return 0x2E;   //unused
                case "switch": return 0x2F;
                case "ENDSWITCH": return 0x30;  //unused
                case "case": return 0x31;
                case "default": return 0x32;
                case "return": return 0x33;
                case "function": return 0x34;
                case "on": return 0x35;
                case "scriptvar": return 0x36;
                case "include": return 0x37;
                case "extern": return 0x38;
                case "~0x3A~": return 0x3A;
                case "~0x3C~": return 0x3C;
                case "~0x3D~": return 0x3D;
                case "~0x3E~": return 0x3E;
                case "~0x41~": return 0x41;
                default: return 0xFF;
            }
        }

        private static short ByteToObfuscated(byte obfus)
        {
            switch (obfus)
            {
                case 0x02: return 0x390C; // (
                case 0x03: return 0x1238; // )
                case 0x04: return 0x7FF5; // ,
                case 0x06: return 0x0732; // ;
                case 0x07: return 0x3A9E; // {
                case 0x08: return 0x3EF6; // }
                case 0x09: return 0x3699; // [
                case 0x0A: return 0x16C5; // ]
                case 0x0B: return 0x5DB2; // ! (not)
                case 0x0C: return 0x2F14; // + (add)
                case 0x0D: return 0x7DD1; // - (substract)
                case 0x0E: return 0x401D; // * (multiply)
                case 0x0F: return 0x2833; // / (divide)
                case 0x10: return 0x127E; // % (modula)
                case 0x11: return 0x01D3; // == (equal to)
                case 0x12: return 0x252A; // != (not equal to)
                case 0x13: return 0x2B01; // < (lower then)
                case 0x14: return 0x0C15; // > (higher then)
                case 0x15: return 0x3492; // <= (lower then or equal to)
                case 0x16: return 0x1796; // >= (higher then or equal to)
                case 0x17: return 0x1316; // = (assign)
                case 0x18: return 0x0DE5; // && (logical and)
                case 0x19: return 0x3960; // || (logical or)
                case 0x1B: return 0x1D11; // ++ (increment)
                case 0x1C: return 0x3F97; // -- (decrement)
                case 0x1D: return 0x3EE9; // int (integer)
                case 0x1E: return 0x2FE7; // str (string)
                case 0x1F: return 0x5C5E; // ust (?)
                case 0x20: return 0x2BA5; // loc (location (x, y, z))
                case 0x21: return 0x36C2; // obj (object)
                case 0x22: return 0x113E; // lis (list)
                case 0x23: return 0x7E0E; // voi (void)
                case 0x25: return 0x1003; // if
                case 0x26: return 0x3308; // else
                case 0x28: return 0x5D2B; // while
                case 0x2A: return 0x1ECA; // for
                case 0x2C: return 0x27D3; // continue
                case 0x2D: return 0x13A6; // break
                case 0x2F: return 0x0914; // switch
                case 0x31: return 0x342D; // case
                case 0x32: return 0x31AD; // default
                case 0x33: return 0x2F15; // return
                case 0x34: return 0x0A2F; // "declares a function and defines the implementation!"
                case 0x35: return 0x038F; // "implement an event"
                case 0x36: return 0x2FA1; // "declare a variable at script level"
                case 0x37: return 0x1B32; // "include another script"
                /// if you include a script it must the very first statement and you can only include one script per script!
                case 0x38: return 0x7FA6; // "declares a function without defining the implementation!"
                case 0x3A: return 0x2BB8; // quoted-string (C style = terminated with a 0-byte)
                case 0x3C: return 0x15E1; // constant byte (gets zero extended to signed 32-bit integer)
                case 0x3D: return 0x3C8A; // constant word (gets zero extended to signed 32-bit integer)
                case 0x3E: return 0x093B; // constant dword (gets casted to signed 32-bit integer)
                case 0x41: return 0x3510; // string (C style = terminated with a 0-byte)
                case 0x1A: return 0x282D; // ^ (xor)
                //This is a list of defined but unsupported tokens:
                case 0x27: return 0x123B; // ENDIF
                case 0x29: return 0x1295; // ENDWHILE
                case 0x2B: return 0x1AF6; // ENDFOR
                case 0x2E: return 0x2CD5; // GOTO (it's unclear how this statement works!)
                case 0x30: return 0x198C; // ENDSWITCH
                //This is a list of obfuscated but unsupported tokens
                case 0x01: return 0x7E87;
                case 0x05: return 0x0BDB;
                case 0x24: return 0x1DCB;
                case 0x39: return 0x03F9;
                case 0x3B: return 0x7B8B;
                case 0x3F: return 0x2410;
                case 0x40: return 0x2BEF;
                default: return 0xFF;
            }
        }
    }
}
