﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace ScriptEditor
{
    public class ScriptDecompile
    {
        public enum NextIs
        {
            None,
            String,
            Byte,
            Short,
            Int32
        }

        public static string Decompile(string filename)
        {
            string output = String.Empty;
            NextIs next = NextIs.None;
            bool casevalue = false;
            bool eventargs = false;
            string eventargsstr = String.Empty;
            bool lastwason = false;
            bool ison = false;
            int tablevel = 0;
            int bracketlevel = 0;

            StreamReader stream = new StreamReader(filename);

            using (BinaryReader br = new BinaryReader(stream.BaseStream))
            {
                while (stream.BaseStream.Length >= stream.BaseStream.Position + 2)
                {
                    short shrt;
                    byte b;
                    switch (next)
                    {
                        case NextIs.None:
                            {
                                shrt = br.ReadInt16();
                                b = ObfuscatedToByte(shrt);
                                if (b == 0x07) // {
                                {
                                    ++tablevel; ison = false;
                                    if (eventargs)
                                    {
                                        eventargs = false;
                                        output += eventargsstr;
                                        eventargsstr = String.Empty;
                                    }
                                    output += "\r\n";
                                }
                                else if (b == 0x08) --tablevel; // }
                                if (b == 0x02) { ++bracketlevel; }  // (
                                else if (b == 0x03) --bracketlevel; // )
                                if (b == 0x31) casevalue = true; // case
                                if (b == 0x35) { eventargs = true; ison = true; lastwason = true; } // on
                                switch (b)
                                {
                                    case 0x41:
                                    case 0x3A: next = NextIs.String; break;
                                    case 0x3C: next = NextIs.Byte; break;
                                    case 0x3D: next = NextIs.Short; break;
                                    case 0x3E: next = NextIs.Int32; break;
                                    default: output += gettab(b, tablevel) + (ison && b == 0x02 ? "<" : ison && b == 0x03 ? ">" : ByteToString(b)); break;
                                }
                                if (b == 0x32) output += ":\r\n";   //default:
                                if (b == 0x06 && bracketlevel == 0) output += "\r\n";
                                break;
                            }
                        case NextIs.String:
                            {
                                shrt = br.ReadInt16();
                                b = ObfuscatedToByte(shrt);
                                next = NextIs.None;
                                if (bracketlevel == 0) output += gettab(0x25, tablevel);
                                output += SDB.FindString((uint)shrt);
                                lastwason = false;
                                if (eventargs && eventargsstr == String.Empty)
                                    eventargsstr = GetEventArgs(SDB.FindString((uint)shrt));
                                if (casevalue)
                                {
                                    casevalue = false;
                                    output += ":\r\n";
                                }
                                break;
                            }
                        case NextIs.Byte:
                            {
                                next = NextIs.None;
                                output += "0x" + br.ReadByte().ToString("X2");
                                if (lastwason)
                                {
                                    output += " ";
                                    lastwason = false;
                                }
                                if (casevalue)
                                {
                                    casevalue = false;
                                    output += ":\r\n";
                                }
                                break;
                            }
                        case NextIs.Short:
                            {
                                next = NextIs.None;
                                 output += "0x" + br.ReadInt16().ToString("X4");
                                if (lastwason)
                                {
                                    output += " ";
                                    lastwason = false;
                                }
                                if (casevalue)
                                {
                                    casevalue = false;
                                    output += ":\r\n";
                                }
                                break;
                            }
                        case NextIs.Int32:
                            {
                                next = NextIs.None;
                                output += "0x" + br.ReadInt32().ToString("X8");
                                if (lastwason)
                                {
                                    output += " ";
                                    lastwason = false;
                                }
                                if (casevalue)
                                {
                                    casevalue = false;
                                    output += ":\r\n";
                                }
                                break;
                            }
                    }
                }
            }

            return output;
        }

        static string gettab(byte b, int level)
        {
            string outp = string.Empty;
            for (int i = 0; i < (b == 0x07 ? level - 1 : level); ++i)
                outp += "    ";
            switch (b)
            {
                case 0x07:
                case 0x08:
                //case 0x1D:
                case 0x1E:
                case 0x1F:
                //case 0x21:
                case 0x22:
                case 0x23:
                case 0x25:
                case 0x26:
                case 0x28:
                case 0x2A:
                case 0x2D:
                case 0x2F:
                case 0x31:
                case 0x32:
                case 0x33:
                case 0x34:
                case 0x35:
                case 0x36:
                case 0x38: return outp;
                default: return string.Empty;
            }
        }

        private static byte ObfuscatedToByte(short obfus)
        {
            switch (obfus)
            {
                case 0x390C:
                case 0x0F3E:
                case 0x0199:
                case 0x0124:
                case 0x305E: return 0x02; // (
                case 0x39B3:
                case 0x2D12:
                case 0x26A6:
                case 0x5D03:
                case 0x1238: return 0x03; // )
                case 0x3B25:
                case 0x1E1F:
                case 0x1AD4:
                case 0x7F96:
                case 0x7FF5: return 0x04; // ,
                case 0x0732:
                case 0x0120:
                case 0x5CFD:
                case 0x3E12:
                case 0x3BF6: return 0x06; // ;
                case 0x3A9E:
                case 0x0DDC:
                case 0x5E14:
                case 0x2E40:
                case 0x1CD0: return 0x07; // {
                case 0x7EB7:
                case 0x6032:
                case 0x2C3B:
                case 0x15A1:
                case 0x3EF6: return 0x08; // }
                case 0x409D:
                case 0x12E1:
                case 0x121F:
                case 0x26CA:
                case 0x3699: return 0x09; // [
                case 0x0902:
                case 0x7BB9:
                case 0x139D:
                case 0x187E:
                case 0x16C5: return 0x0A; // ]
                case 0x3CD5:
                case 0x13E9:
                case 0x4080:
                case 0x5DB2:
                case 0x33EA: return 0x0B; // ! (not)
                case 0x23C9:
                case 0x60BF:
                case 0x3CD6:
                case 0x0FBF:
                case 0x2F14: return 0x0C; // + (add)
                case 0x047E:
                case 0x368E:
                case 0x2FFF:
                case 0x288F:
                case 0x7DD1: return 0x0D; // - (substract)
                case 0x261E:
                case 0x5E9D:
                case 0x1916:
                case 0x32E6:
                case 0x401D: return 0x0E; // * (multiply)
                case 0x0384:
                case 0x18D7:
                case 0x0FC9:
                case 0x0E12:
                case 0x2833: return 0x0F; // / (divide)
                case 0x249E:
                case 0x2B0C:
                case 0x11F4:
                case 0x5DD5:
                case 0x127E: return 0x10; // % (modula)
                case 0x0135:
                case 0x07CF:
                case 0x1AF4:
                case 0x0ECC:
                case 0x01D3: return 0x11; // == (equal to)
                case 0x0E90:
                case 0x3A2D:
                case 0x37E6:
                case 0x19D9:
                case 0x252A: return 0x12; // != (not equal to)
                case 0x37E5:
                case 0x1DC0:
                case 0x1481:
                case 0x4087:
                case 0x2B01: return 0x13; // < (lower then)
                case 0x16D4:
                case 0x3A8D:
                case 0x7FBE:
                case 0x0C7B:
                case 0x0C15: return 0x14; // > (higher then)
                case 0x3807:
                case 0x0633:
                case 0x251F:
                case 0x1D18:
                case 0x3492: return 0x15; // <= (lower then or equal to)
                case 0x19DA:
                case 0x39CE:
                case 0x3BB1:
                case 0x3004:
                case 0x1796: return 0x16; // >= (higher then or equal to)
                case 0x1F16:
                case 0x182F:
                case 0x2CF7:
                case 0x5ED0:
                case 0x1316: return 0x17; // = (assign)
                case 0x5D24:
                case 0x0588:
                case 0x7CFE:
                case 0x2725:
                case 0x0DE5: return 0x18; // && (logical and)
                case 0x13D3:
                case 0x29D8:
                case 0x0A28:
                case 0x09CE:
                case 0x3960: return 0x19; // || (logical or)
                case 0x5CCD:
                case 0x0940:
                case 0x293B:
                case 0x40A5:
                case 0x1D11: return 0x1B; // ++ (increment)
                case 0x2528:
                case 0x0EA9:
                case 0x3F0B:
                case 0x3087:
                case 0x3F97: return 0x1C; // -- (decrement)
                case 0x30F1:
                case 0x3295:
                case 0x01C1:
                case 0x0CE1:
                case 0x3EE9: return 0x1D; // intege
                case 0x3F9A:
                case 0x30A7:
                case 0x2DB5:
                case 0x169A:
                case 0x2FE7: return 0x1E; // string
                case 0x10D9:
                case 0x0390:
                case 0x2A38:
                case 0x0728:
                case 0x5C5E: return 0x1F; // ust (?)
                case 0x01E1:
                case 0x1030:
                case 0x1BD9:
                case 0x159F:
                case 0x2BA5: return 0x20; // location (location (x, y, z))
                case 0x28E2:
                case 0x2F0C:
                case 0x1289:
                case 0x3382:
                case 0x36C2: return 0x21; // object
                case 0x26B1:
                case 0x1CDF:
                case 0x27DA:
                case 0x0E29:
                case 0x113E: return 0x22; // list
                case 0x2E39:
                case 0x1D3F:
                case 0x1D5E:
                case 0x1FF1:
                case 0x7E0E: return 0x23; // void
                case 0x12C2:
                case 0x1003:
                case 0x0607:
                case 0x0784:
                case 0x2B0F: return 0x25; // if
                case 0x3305:
                case 0x32E7:
                case 0x212C:
                case 0x018E:
                case 0x3308: return 0x26; // else
                case 0x3106:
                case 0x018C:
                case 0x357E:
                case 0x0A87:
                case 0x5D2B: return 0x28; // while
                case 0x7DAA:
                case 0x2F0B:
                case 0x1BFC:
                case 0x13F5:
                case 0x1ECA: return 0x2A; // for
                case 0x017B:
                case 0x6014:
                case 0x0E99:
                case 0x33CD:
                case 0x27D3: return 0x2C; // continue
                case 0x7F0D:
                case 0x04F0:
                case 0x183A:
                case 0x1FB4:
                case 0x13A6: return 0x2D; // break
                case 0x04B0:
                case 0x1927:
                case 0x08FF:
                case 0x31D8:
                case 0x0914: return 0x2F; // switch
                case 0x3223:
                case 0x17B8:
                case 0x3895:
                case 0x248D:
                case 0x342D: return 0x31; // case
                case 0x5D3D:
                case 0x3260:
                case 0x32DE:
                case 0x2780:
                case 0x31AD: return 0x32; // default
                case 0x5DE9:
                case 0x5EA5:
                case 0x11D5:
                case 0x199F:
                case 0x2F15: return 0x33; // return
                case 0x0E01:
                case 0x19FE:
                case 0x3821:
                case 0x0B93:
                case 0x0A2F: return 0x34; // "declares a function and defines the implementation!"
                case 0x09B3:
                case 0x038F:
                case 0x328A:
                case 0x08AF:
                case 0x5CCA: return 0x35; // "implement an event"
                case 0x0C95:
                case 0x7CBE:
                case 0x7C27:
                case 0x5D2A:
                case 0x2FA1: return 0x36; // "declare a variable at script level"
                case 0x31BE:
                case 0x15B4:
                case 0x07C9:
                case 0x27C0:
                case 0x1B32: return 0x37; // "include another script"
                /// if you include a script it must the very first statement and you can only include one script per script!
                case 0x2934:
                case 0x3E09:
                case 0x012C:
                case 0x2CC6:
                case 0x7FA6: return 0x38; // "declares a function without defining the implementation!"
                case 0x5D17:
                case 0x0A1D:
                case 0x3B29:
                case 0x2BFA:
                case 0x2BB8: return 0x3A; // quoted-string (C style = terminated with a 0-byte)
                case 0x17BD:
                case 0x21EB:
                case 0x2015:
                case 0x5DB8:
                case 0x15E1: return 0x3C; // constant byte (gets zero extended to signed 32-bit integer)
                case 0x5B60:
                case 0x3D8F:
                case 0x0FF4:
                case 0x275B:
                case 0x3C8A: return 0x3D; // constant word (gets zero extended to signed 32-bit integer)
                case 0x188F:
                case 0x5D27:
                case 0x7F5C:
                case 0x01F7:
                case 0x093B: return 0x3E; // constant dword (gets casted to signed 32-bit integer)
                case 0x3510:
                case 0x0B9B:
                case 0x06E9:
                case 0x3B9E:
                case 0x0B31: return 0x41; // string (C style = terminated with a 0-byte)
                case 0x263D:
                case 0x3B97:
                case 0x4027:
                case 0x138A:
                case 0x282D: return 0x1A; // ^ (xor)
                //This is a list of defined but unsupported tokens:
                case 0x1EDC:
                case 0x20A8:
                case 0x37BE:
                case 0x01EB:
                case 0x123B: return 0x27; // ENDIF
                case 0x03FA:
                case 0x0AF0:
                case 0x0786:
                case 0x2332:
                case 0x1295: return 0x29; // ENDWHILE
                case 0x0D9F:
                case 0x388A:
                case 0x15FD:
                case 0x7CB8:
                case 0x1AF6: return 0x2B; // ENDFOR
                case 0x190B:
                case 0x3605:
                case 0x20AD:
                case 0x32CF:
                case 0x2CD5: return 0x2E; // GOTO (it's unclear how this statement works!)
                case 0x13F4:
                case 0x3A27:
                case 0x387C:
                case 0x32C1:
                case 0x198C: return 0x30; // ENDSWITCH
                //This is a list of obfuscated but unsupported tokens
                case 0x0BB3:
                case 0x2EA6:
                case 0x12DB:
                case 0x153C:
                case 0x7E87: return 0x01;
                case 0x323B:
                case 0x260D:
                case 0x030A:
                case 0x301C:
                case 0x0BDB: return 0x05;
                case 0x06E3:
                case 0x36A1:
                case 0x0C1E:
                case 0x2120:
                case 0x1DCB: return 0x24;
                case 0x06D8:
                case 0x3181:
                case 0x2738:
                case 0x3212:
                case 0x03F9: return 0x39;
                case 0x1DB5:
                case 0x2F95:
                case 0x0EF5:
                case 0x5CDF:
                case 0x7B8B: return 0x3B;
                case 0x2F84:
                case 0x1DC3:
                case 0x1DA7:
                case 0x1A30:
                case 0x2410: return 0x3F;
                case 0x02EE:
                case 0x0603:
                case 0x012F:
                case 0x2C9E:
                case 0x2BEF: return 0x40;
                default: return 0xFF;
            }
        }

        private static string ByteToString(byte obfusbyte)
        {
            switch (obfusbyte)
            {
                case 0x02: return "(";
                case 0x03: return ")";
                case 0x04: return ", ";
                case 0x06: return ";";
                case 0x07: return "{\r\n";
                case 0x08: return "}\r\n";
                case 0x09: return "[";
                case 0x0A: return "]";
                case 0x0B: return "!";
                case 0x0C: return "+";
                case 0x0D: return "-";
                case 0x0E: return " * ";
                case 0x0F: return " / ";
                case 0x10: return " % ";
                case 0x11: return " == ";
                case 0x12: return " != ";
                case 0x13: return " < ";
                case 0x14: return " > ";
                case 0x15: return " <= ";
                case 0x16: return " >= ";
                case 0x17: return " = ";
                case 0x18: return " && ";
                case 0x19: return " || ";
                case 0x1A: return " ^ "; //(xor)
                case 0x1B: return "++";
                case 0x1C: return "--";
                case 0x1D: return "integer ";
                case 0x1E: return "string ";
                case 0x1F: return "ust ";
                case 0x20: return "location "; // location
                case 0x21: return "object ";
                case 0x22: return "list ";
                case 0x23: return "void ";
                case 0x25: return "if ";
                case 0x26: return "else ";
                case 0x27: return "ENDIF";  //unused
                case 0x28: return "while ";
                case 0x29: return "ENDWHILE";   //unused
                case 0x2A: return "for ";
                case 0x2B: return "ENDFOR";
                case 0x2C: return "continue";
                case 0x2D: return "break";
                case 0x2E: return "GOTO";   //unused
                case 0x2F: return "switch ";
                case 0x30: return "ENDSWITCH";  //unused
                case 0x31: return "case ";
                case 0x32: return "default";
                case 0x33: return "return";
                case 0x34: return "function ";
                case 0x35: return "on ";
                case 0x36: return "scriptvar ";
                case 0x37: return "include ";
                case 0x38: return "extern ";
                case 0x3A: return "~0x3A~";
                case 0x3C: return "~0x3C~";
                case 0x3D: return "~0x3D~";
                case 0x3E: return "~0x3E~";
                case 0x41: return "~0x41~";
                default: return "UNK";
            }
        }

        private static string GetEventArgs(string eventname)
        {
            switch (eventname.ToLower())
            {
                case "speech": return "(object speaker, string arg)";
                case "gotattacked":
                case "killedtarget":
                case "minrangedefend":
                case "maxrangedefend": return "(object attacker)";
                case "sawdeath": return "(object attacker, object victim, object corpse)";
                case "message": return "(object sender, list args)";
                case "use":
                case "ooruse": return "(object user)";
                case "washit": return "(object attacker, integer damamt)";
                case "targetloc": return "(object user, location place, integer objtype)";
                case "wasdropped": return "(object dropper)";
                case "lookedat": return "(object looker)";
                case "give": return "(object giver, object givenobj)";
                case "wasgotten": return "(object getter)";
                case "ishitting": return "(object victim, integer damamt)";
                case "convofunc": return "(object talker, string arg)";
                case "typeselected": return "(object user, integer listindex, integer objtype, integer objhue)";
                case "hueselected": return "(object user, integer objhue)";
                case "moon": return "(integer trammelchange, integer feluccachange)";
                case "minrangeattack":
                case "maxrangeattack": return "(object defender)";
                case "equip": return "(object equippedon)";
                case "unequip": return "(object unequippedfrom)";
                case "isstackableon":
                case "stackonto": return "(object stackon)";
                case "multirecycle": return "(integer oldtype, integer newtype)";
                case "decay": return "(integer oldvalue, integer newvalue)";
                case "genericgump": return "(object user, integer closeId, list selectList, list entryList)";
                case "pkpost": return "(object killer, object killee)";
                case "textentry": return "(object sender, integer button, string text)";
                case "shop": return "(integer func)";
                case "stolenfrom": return "(object stealer)";
                case "objaccess": return "(object user, object usedon)";
                case "transaccountcheck":
                case "transresponse": return "(object target, integer transok)";
                case "canbuy": return "(object buyer, object seller, integer quantity)";
                case "mobishitting": return "(object victim, integer damage)";
                case "aversion":
                case "fightpulse":
                case "foundfood":
                case "founddesire":
                case "foundshelter":
                case "acquiredesire":
                case "enterrange":
                case "leaverange": return "(object target)";
                case "targetobj":
                case "oortargetobj": return "(object user, object usedon)";
                case "creation":
                case "death":
                case "failfood":
                case "faildesire":
                case "failshelter":
                case "time":
                case "loiter":
                case "seekfood":
                case "seekdesire":
                case "seekshelter":
                case "weather":
                case "pathfound":
                case "pathnotfound":
                case "callback":
                case "destroyed":
                case "serverswitch":
                case "logout":
                case "objectloaded":
                case "ishealthy":
                case "online":
                case "famechanged":
                case "karmachanged":
                case "murdercountchanged":
                default: return "()";
            }
        }
    }
}
