﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace ScriptEditor
{
    public class SDB
    {
        private static string m_sdbfilepath;
        private static uint m_Lines = 0;
        private static List<SDBEntry> m_Entries = new List<SDBEntry>();
        private static bool m_Initialized = false;
        public static bool IsInitialized { get { return m_Initialized; } }
        public static uint Lines { get { return m_Lines; } }

        public static void InitSDB(string filepath)
        {
            m_sdbfilepath = filepath;
            m_Entries.Clear();
            m_Lines = 0;
            using (StreamReader sr = new StreamReader(m_sdbfilepath))
            {
                while (!sr.EndOfStream)
                {
                    m_Entries.Add(new SDBEntry(m_Lines, sr.ReadLine()));
                    ++m_Lines;
                }
            }
            m_Initialized = true;
        }

        public static uint FindLine(string text)
        {
            if (IsInitialized)
            {
                SDBEntry res = m_Entries.Find(delegate(SDBEntry entry) { return entry.Content.Equals(text); });
                if (res.IsEmpty)
                    return 0;
                else
                    return res.Line;
            }
            else
                throw new Exception("SDB is not initialized.\r\nUse InitSDB(string filepath) first.");
        }

        public static string FindString(uint line)
        {
            if (IsInitialized)
            {
                SDBEntry res = m_Entries.Find(delegate(SDBEntry entry) { return entry.Line.Equals(line); });
                if (res.IsEmpty)
                    return null;
                else
                    return res.Content;
            }
            else
                throw new Exception("SDB is not initialized.\r\nUse InitSDB(string filepath) first.");
        }

        public static uint WriteEntry(string text)
        {
            string tmp = text.Replace("\\", "\\\\");
            tmp = tmp.Replace("\r", "\\r");
            tmp = tmp.Replace("\n", "\\n");
            using (StreamWriter sw = new StreamWriter(m_sdbfilepath, true))
            {
                sw.WriteLine(tmp);
            }
            m_Entries.Add(new SDBEntry(m_Lines, tmp));
            return m_Lines++;
        }

        public struct SDBEntry
        {
            uint m_Line;
            string m_Content;
            public uint Line { get { return m_Line; } }
            public string Content { get { return m_Content; } }
            public bool IsEmpty { get { return (m_Line == 0 && m_Content == null); } }
            public SDBEntry(uint line, string content)
            {
                m_Line = line;
                m_Content = content;
            }
            public override string ToString()
            {
                return m_Content;
            }
        }
    }
}
